
/* Bit array support

   Copyright 2016 Molnár Károly
*/

#include "libgputils.h"

#define BITARRAY_GROUP_SIZE             (sizeof(uint64_t) * 8)

/*------------------------------------------------------------------------------------------------*/

void
gp_bitarray_create(gp_bit_array_t *Bits, size_t Bit_size)
{
  assert(Bits);

  if (Bit_size == 0) {
    Bits->bit_length   = 0;
    Bits->group_length = 0;
    Bits->byte_length  = 0;
    Bits->array        = NULL;
  }

  Bits->bit_length   = Bit_size;
  Bits->group_length = (Bit_size + BITARRAY_GROUP_SIZE - 1) / BITARRAY_GROUP_SIZE;
  Bits->byte_length  = Bits->group_length * (sizeof(uint64_t) / sizeof(uint8_t));
  Bits->array        = GP_Calloc(sizeof(uint64_t), Bits->group_length);
}

/*------------------------------------------------------------------------------------------------*/

void
gp_bitarray_delete(gp_bit_array_t *Bits)
{
  assert(Bits);

  if (Bits->array) {
    free(Bits->array);
    Bits->array = NULL;
  }

  Bits->bit_length   = 0;
  Bits->byte_length  = 0;
  Bits->group_length = 0;
}

/*------------------------------------------------------------------------------------------------*/

bool
gp_bitarray_write(gp_bit_array_t *Bits, size_t Bit_index, bool Value)
{
  size_t   group_index;
  size_t   bit_index;
  uint64_t cluster;

  assert(Bits);

  if ((Bits->array == NULL) || (Bit_index >= Bits->bit_length)) {
    return false;
  }

  group_index = Bit_index / BITARRAY_GROUP_SIZE;
  bit_index   = Bit_index % BITARRAY_GROUP_SIZE;
  cluster     = Bits->array[group_index];

  if (Value) {
    cluster |= 1 << bit_index;
  }
  else {
    cluster &= ~(1 << bit_index);
  }

  Bits->array[group_index] = cluster;
  return true;
}

/*------------------------------------------------------------------------------------------------*/

bool
gp_bitarray_write_range(gp_bit_array_t *Bits, size_t Bit_index_start, size_t Bit_index_end,
                        bool Value)
{
  size_t   group_index_start;
  size_t   bit_index_start;
  size_t   group_index_end;
  size_t   bit_index_end;
  size_t   i;
  uint64_t mask_start;
  uint64_t mask_end;
  uint64_t cluster;

  assert(Bits);

  if ((Bits->array == NULL) || (Bit_index_end >= Bits->bit_length)) {
    return false;
  }

  if (Bit_index_start > Bit_index_end) {
    i               = Bit_index_end;
    Bit_index_end   = Bit_index_start;
    Bit_index_start = i;
  }

  group_index_start = Bit_index_start / BITARRAY_GROUP_SIZE;
  bit_index_start   = Bit_index_start % BITARRAY_GROUP_SIZE;
  group_index_end   = Bit_index_end   / BITARRAY_GROUP_SIZE;
  bit_index_end     = Bit_index_end   % BITARRAY_GROUP_SIZE;

  mask_start = (uint64_t)(-1);
  mask_end   = (uint64_t)(-1);
  mask_start <<= bit_index_start;
  mask_end   >>= BITARRAY_GROUP_SIZE - bit_index_end - 1;

  i = group_index_end - group_index_start;
  cluster = Bits->array[group_index_start];

  if (i == 0) {
    mask_start &= mask_end;

    if (Value) {
      cluster |= mask_start;
    }
    else {
      cluster &= ~(mask_start);
    }

    Bits->array[group_index_start] = cluster;
    return true;
  }

  if (Value) {
    cluster |= mask_start;
  }
  else {
    cluster &= ~(mask_start);
  }

  Bits->array[group_index_start] = cluster;

  if (i > 1) {
    cluster = (Value) ? (uint64_t)(-1) : 0;
    for (i = group_index_start + 1; i < group_index_end; ++i) {
      Bits->array[i] = cluster;
    }
  }

  cluster = Bits->array[group_index_end];

  if (Value) {
    cluster |= mask_end;
  }
  else {
    cluster &= ~(mask_end);
  }

  Bits->array[group_index_end] = cluster;
  return true;
}

/*------------------------------------------------------------------------------------------------*/

bool
gp_bitarray_clear(gp_bit_array_t *Bits)
{
  assert(Bits);

  if (Bits->array == NULL) {
    return false;
  }

  memset(Bits->array, 0, Bits->byte_length);
  return true;
}

/*------------------------------------------------------------------------------------------------*/

bool
gp_bitarray_read(const gp_bit_array_t *Bits, size_t Bit_index)
{
  size_t   group_index;
  size_t   bit_index;
  uint64_t cluster;

  assert(Bits);

  if ((Bits->array == NULL) || (Bit_index >= Bits->bit_length)) {
    return false;
  }

  group_index = Bit_index / BITARRAY_GROUP_SIZE;
  bit_index   = Bit_index % BITARRAY_GROUP_SIZE;
  cluster     = Bits->array[group_index];

  return ((cluster & (1 << bit_index)) ? true : false);
}

/*------------------------------------------------------------------------------------------------*/

static bool
_find_lowest_bit(const uint64_t *Array, size_t Group_index, size_t *Find)
  {
  bool first;
  size_t     num;
  uint64_t   cluster;

  first = true;
  num   = 0;
  while (true) {
    cluster = Array[Group_index];
    if (cluster == 0) {
      if (first) {
        return false;
      }

      ++Group_index;
      *Find = num - 1 + (Group_index * BITARRAY_GROUP_SIZE);
      return true;
    }

    num = gp_find_lowest_bit(cluster);
    if (num > 1) {
      *Find = num - 1 + (Group_index * BITARRAY_GROUP_SIZE);
      return true;
    }

    if (Group_index == 0) {
      break;
    }

    --Group_index;
    first = false;
  }

  return false;
}

/*------------------------------------------------------------------------------------------------*/

static bool
_find_highest_bit(const uint64_t *Array, size_t Group_index, size_t Group_end, size_t *Find)
  {
  bool first;
  size_t     num;
  uint64_t   cluster;

  first = true;
  num   = 0;
  while (true) {
    cluster = Array[Group_index];
    if (cluster == 0) {
      if (first) {
        return false;
      }

      --Group_index;
      *Find = num - 1 + (Group_index * BITARRAY_GROUP_SIZE);
      return true;
    }

    num = gp_find_highest_bit(cluster);
    if (num < 64) {
      *Find = num - 1 + (Group_index * BITARRAY_GROUP_SIZE);
      return true;
    }

    if (Group_index >= Group_end) {
      break;
    }

    ++Group_index;
    first = false;
  }

  return false;
}

/*------------------------------------------------------------------------------------------------*/

bool
gp_bitarray_get_range_borders(const gp_bit_array_t *Bits, size_t Bit_index, size_t *Start, size_t *End)
  {
  size_t     group_index;
  bool ret;

  assert(Bits);
  assert(Start);
  assert(End);

  if ((Bits->array == NULL) || (Bit_index >= Bits->bit_length)) {
    return false;
  }

  group_index = Bit_index / BITARRAY_GROUP_SIZE;
  ret  = _find_lowest_bit(Bits->array, group_index, Start);
  ret |= _find_highest_bit(Bits->array, group_index, Bits->group_length, End);
  return ret;
}
